import numpy as np
from dasy.core import Labeler

class CentroidsLabeler(Labeler):
    """
    Creates a centroid for each class and labels points according to 
    which is the closest centroid (similar to the first step of k-means).
    This is a deterministic labeler. 
    """
    def __init__(self, classes=2, dim=2):
        super().__init__(classes=classes)
        self.dim = dim
        self.centroids = np.random.normal(size=(classes, dim))

    def assign(self, X):
        y = np.zeros(len(X))
        for i in range(len(X)):
            y[i] = self.nearest_centroid(X[i], self.centroids)
        return y

    def nearest_centroid(self, x, centroids):
        dists = np.linalg.norm(x - centroids, axis=1)
        return np.argmin(dists)




        



import numpy as np
from dasy.core import DataSynthesizer

#TODO: switch from numpy random to Generator

class GaussianSynth(DataSynthesizer):
    """
    Samples from a single multivariate Normal Distribution
    """
    def sample(self, n=100, dim=2, mean=None, cov=None):
        """
        See numpy.random.multivariate_normal for details on params
        We add simple defaults for mean and cov
        """
        if mean is None:
            mean = np.zeros(dim)
        if cov is None:
            cov = np.eye(dim)
        return np.random.multivariate_normal(mean=mean, cov=cov, size=n)




import numpy as np

class DataSynthesizer:
    """
    Samples input X data according to a specified process.
    """
    def __init__(self):
        pass

    def sample(self, n=100, dim=2):
        """
        Returns a sample of synthetic data. 
        @param n: Number of data points to sample
        @param dim: Dimension of data
        @return data X
        """
        raise NotImplementedError

class Labeler:
    """
    Assigns labels y to input X data. 
    """
    def __init__(self, classes):
        self.classes = classes

    def assign(self, X):
        """
        Assign labels y to input X data
        @return labels y, aligned with X
        """
        raise NotImplementedError




import numpy as np
import matplotlib.pyplot as plt
from synthesizers.gaussian import GaussianSynth
from labelers.centroids import CentroidsLabeler

def main():
    dim = 2
    #plt.gca().set_aspect('equal', adjustable='box')
    #fig, ax = plt.subplots()
    f, (ax1, ax2) = plt.subplots(1, 2, sharey=True)
    f.suptitle('Gaussian Data with Centroids Labeling', y=0.84)
    ax1.set_aspect('equal', adjustable='box')
    ax2.set_aspect('equal', adjustable='box')
    ax1.set_xlim(-2, 2)
    ax1.set_ylim(-2, 2)
    ax2.set_xlim(-2, 2)
    ax2.set_ylim(-2, 2)

    synth = GaussianSynth()
    data = synth.sample(n=100, dim=dim)
    #plt.scatter(data.T[0], data.T[1])
    #plt.show()

    labeler = CentroidsLabeler(classes=2, dim=dim)
    labels = labeler.assign(data)
    ax1.scatter(data.T[0], data.T[1], c=labels)
    ax1.scatter(labeler.centroids.T[0], labeler.centroids.T[1], c='r')
    ax1.plot(labeler.centroids.T[0], labeler.centroids.T[1], '--')
    plot_bisectors(labeler.centroids[0], labeler.centroids[1], ax=ax1)
    ax1.title.set_text('classes=2')
    #plt.show()

    labeler = CentroidsLabeler(classes=3, dim=dim)
    labels = labeler.assign(data)
    ax2.scatter(data.T[0], data.T[1], c=labels)
    ax2.scatter(labeler.centroids.T[0], labeler.centroids.T[1], c='r')
    #plt.plot(labeler.centroids.T[0], labeler.centroids.T[1], '--')
    plot_bisectors(labeler.centroids[0], labeler.centroids[1], ax=ax2)
    plot_bisectors(labeler.centroids[0], labeler.centroids[2], ax=ax2)
    plot_bisectors(labeler.centroids[1], labeler.centroids[2], ax=ax2)
    ax2.title.set_text('classes=3')
    plt.tight_layout()
    plt.show()

def plot_bisectors(a, b, ax=plt):
    midpoint = (a + b) / 2
    slope = - (b[0] - a[0]) / (b[1] - a[1])
    #plt.plot([midpoint[0], midpoint[0] + 1], [midpoint[1], midpoint[1] + slope], '-')
    ax.axline((midpoint[0], midpoint[1]), slope=slope, linestyle='--')

if __name__ == '__main__':
    main()

